#!/usr/bin/env python
"""
Functions to filter geodynamic models

These are used to provide a command line interface to
common filtering tasks.
"""
from pathlib import Path
from typing_extensions import Annotated
from dataclasses import dataclass

import typer

from pytomofilt import model


@dataclass
class TomographicModelFiles:
    """Input files for a tomographic model"""
    name: str
    base_path: Path
    coef_file: Path
    weights_file: Path
    evec_file: Path


def tomographic_model_from_path(input_path: Path) -> TomographicModelFiles:
    """
    tomographic_model_from_path builds and validates a tomographic model exists in input_path
    
    :param input_path: Location of tomographic files (directory)
    :type input_path: Path
    :return: Paths to each file needed for tomographic filtering
    :rtype: TomographicModelFiles
    """
    assert input_path.is_dir(), "Input path must be a directory"
    model_name = input_path.parts[-1] # Last part should be a model name
    coef_file = input_path / (model_name + ".sph")
    assert coef_file.is_file(), "Cannot see model coefficient file"
    weights_file = input_path / (model_name + ".smthp_21")
    assert weights_file.is_file(), "Cannot see model weights file"
    evec_file = input_path / (model_name + ".evc")
    assert evec_file.is_file(), "Cannot see model eigenvector file"
    return(TomographicModelFiles(
        name=model_name, base_path=input_path, coef_file=coef_file, weights_file=weights_file,
        evec_file=evec_file)
        )


def ptf_reparam_filter_files(
        tomographic_model: Annotated[Path, typer.Argument(help="Path to a directory containing a tomographic model and filter")],
        geodynamic_model: Annotated[Path, typer.Argument(help="Path to a directory containing data from a geodynamic model")]
    ):
    """
    Reparameterize and filter a geodynamic model so it can be compared with tomography
    """
    # Find the files we will need:
    tomographic_model_spec = tomographic_model_from_path(tomographic_model)
    # Build reference (e.g. tomographic) reference
    print(f"Setting up reference model for {tomographic_model_spec.name}.")
    ref_model = model.RTS_Model.from_file(tomographic_model_spec.coef_file)
    print(f"Setting up filter model for {tomographic_model_spec.name}.")
    ref_model.filter_from_file(tomographic_model_spec.evec_file, tomographic_model_spec.weights_file,
                                0.2,verbose=True)

    # Build a comparison model with the same parameterisation as the 
    # reference model
    print(f"Reading geodynamics model from {geodynamic_model}")
    comp_model = model.RTS_Model.from_directory(geodynamic_model,
            lmax=ref_model.lmax, rmin=ref_model.rmin, rmax=ref_model.rmax,
            knots=ref_model.knots_r)
    comp_model.write(geodynamic_model.parts[-1] + "_reparam.sph")
    # Apply the resolution filter from the reference model to the comparison model
    print(f"Filtering!")
    filtered_comp_model = ref_model.filter(comp_model) 
    filtered_comp_model.write(geodynamic_model.parts[-1] + "_filtered.sph")
    print(f"Done")


if __name__ == "__main__":
    # Just for easy testing - we can loose this later.
    ptf_reparam_filter_files(Path("/Volumes/Elements/LVS_2020_10_29/Code/tomofit_from_Paula/tomofilt_new_ES/utils/S12RTS"), 
                             Path("/Volumes/Elements/LVS_2020_10_29/Code/tomofit_from_Paula/tomofilt_new_ES/geodyn/examplemodel"))


import numpy as np
import pyshtools as shtools

def rts_to_sh(rts_coefs):
    """
    Convert between the SH format from the 'sph' file to pyshtools format

    In '.sph' files used for the SP12RTS tomography model (for example) the 
    spherical harmonic coefficents are complex, fully normalised, with the CS
    phase. For most of our use we use real fully normalised real coefficents 
    without the CS phase. This function does the conversion from 'sph'/'rts' to 
    our normal format.
    """

    inp_shape = np.shape(rts_coefs)
    assert inp_shape[0] == 2, 'rts_coefs must have real and imag parts on 1st dim'
    lmax = inp_shape[1] - 1
    assert inp_shape[2] == lmax + 1, 'must have all ms'

    coefs= shtools.SHCoeffs.from_zeros(lmax, kind='complex', 
                                       normalization='ortho', csphase=-1)

    # We don't need to set -m coeffs. minus sign for imaginary part 
    # because real coefficients for sin store negative m degrees,
    # where sin(-m*phi) = -sin(m*phi)
    coefs.coeffs[0] = rts_coefs[0] - rts_coefs[1] * 1j

    sh_coefs = coefs.convert(normalization='ortho', csphase=1, kind='real', 
                             check=False).to_array()

     # RTS format multiples non-zero order (m) components by 2
    sh_coefs[:,1:,1:] /= 2

    return sh_coefs


def sh_to_rts(sh_coefs):
    """
    Convert between pyshtools format and that from from the 'sph' files

    In '.sph' files used for the SP12RTS tomography model (for example) the 
    spherical harmonic coefficents are complex, fully normalised, with the CS
    phase. For moust of our use we use real fully normalised real coefficents 
    without the CS phase. This function does the conversion from our nornal 
    format to the 'sph'/'rts' format.
    """

    inp_shape = np.shape(sh_coefs)
    assert inp_shape[0] == 2, 'sh_coefs must have real and imag parts on 1st dim'
    lmax = inp_shape[1] - 1
    assert inp_shape[2] == lmax + 1, 'must have all ms'

    real_coefs = shtools.SHCoeffs.from_zeros(lmax, kind='real', 
                                             normalization='ortho', csphase=1)
    
    real_coefs.coeffs = sh_coefs
    sh_coefs[1] = -sh_coefs[1]  # Minus sign for imaginary part because real coefficients for 
                        # sin store negative m degrees, where sin(-m*phi) = -sin(m*phi)

    complex_coefs = real_coefs.convert(normalization='ortho', csphase=-1, kind='complex', 
                              check=False).to_array()

    # SHTOOLS stores the two arrays in axis 0 as positive m and negative m shells. We 
    # want to match the RTS format of storing real part in the first array and
    # imaginary aprt in the second array of axis 0
    real = complex_coefs[0].real
    imag = complex_coefs[0].imag

    rts_coefs = np.array([real, imag])

    # RTS format multiples non-zero order (m) components by 2 and 
    rts_coefs[:,1:,1:] *= 2

    return rts_coefs


import unittest
from unittest.mock import patch, MagicMock
import numpy as np
import numpy.testing as npt

from pytomofilt.filter import Filter

class TestReadEigenvec(unittest.TestCase):

    def setUp(self):
        
        # Setup a dummy Filter object with required attributes
        self.filt = Filter.__new__(Filter)
        self.filt.damp = 0.1
        self.filt.verbose = False

    
    def test_read_eigenvec_header_size_mismatch(self):
        """
        Test assertion that natd == numatd2. numatd2 is
        the second entry of header 1 and natd = (lmaxh+1)**2,
        where lmaxh is the first entry of header 1.        
        """
        mock_file = MagicMock()
        mock_file.read_ints.side_effect = [
            np.array([2, 4, 1, 1, 0, 0, 1]),  # Header 1
            np.array([1, 0, 0, 0, 0, 0, 0])   # Header 2
        ]

        with patch('pytomofilt.filter.spio.FortranFile', return_value=mock_file):
            with self.assertRaises(AssertionError):
                self.filt.read_eigenvec_file("dummy")  


    def test_read_eigenvec_file(self):

        mock_file = MagicMock()
        icrust = 1
        ismth = 1
        lmaxh = 1
        ndep = 1
        eigvals = [1.0, 0.4, 0.3, 0.2] 
        eigvecs = [[1.0, 0.0, 0.0, 0.0],   # First eigenvector
                   [0.0, 1.0, 0.0, 0.0],   # Second eigenvector
                   [0.0, 0.0, 1.0, 0.0],   # Third eigenvector
                   [0.0, 0.0, 0.0, 1.0]]   # Fourth eigenvector
        mock_file.read_ints.side_effect = [
            np.array([lmaxh, 4, ndep, icrust, 0, 0, ismth]),  # Header 1
            np.array([1, 0, 0, 0, 0, 0, 0])   # Header 2
        ]
        mock_file.read_reals.side_effect = [np.array([val]+vec) for val,vec in zip(eigvals,eigvecs)]

        with patch('pytomofilt.filter.spio.FortranFile', return_value=mock_file):
            result = self.filt.read_eigenvec_file("dummy")
            npt.assert_array_equal(eigvals, result[0])
            npt.assert_array_equal(eigvecs, result[1])
            self.assertEqual(icrust, result[2])
            self.assertEqual(ismth, result[3])
            self.assertEqual((lmaxh+1)**2, result[4])
            self.assertEqual(ndep, result[5])


    def test_eigenvec_file_small_eigvals(self):
        """
        Tests that file reading stops after the first eigenvalue smaller 
        than stpfact. Stpfact is given by: largest_eigval * damp / 5000.
        In this example, stpfact = 1.0 * 0.1 / 5000 = 0.002, so the last
        eigenvector should not be recorded.
        """
        mock_file = MagicMock()
        eigvals = [1.0, 0.4, 0.000015, 0.00001] 
        eigvecs = [[1.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 0.0, 0.0],
                   [0.0, 0.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 1.0]]
        mock_file.read_ints.side_effect = [
            np.array([1, 4, 1, 1, 0, 0, 1]),  # Header 1
            np.array([1, 0, 0, 0, 0, 0, 0])   # Header 2
        ]
        mock_file.read_reals.side_effect = [np.array([val]+vec) for val,vec in zip(eigvals,eigvecs)]

        with patch('pytomofilt.filter.spio.FortranFile', return_value=mock_file):
            result = self.filt.read_eigenvec_file("dummy")
            npt.assert_array_equal(eigvals[:-1], result[0])
            npt.assert_array_equal(eigvecs[:-1], result[1])


class TestReadWgts(unittest.TestCase):

    def setUp(self):
        
        # Setup a dummy Filter object with required attributes
        self.filt = Filter.__new__(Filter)
        self.filt.ndep = 3
        self.filt.natd = 9

        self.mock_file = MagicMock()
        self.mock_file.read_ints.return_value = [2, 0, 0, 3, 0, 0, 0, 0, 0]
        self.mock_file.read_reals.return_value = np.arange(0, 9, dtype='float32')


    def test_read_wgts_file(self):

        with patch('pytomofilt.filter.spio.FortranFile', return_value=self.mock_file):
            result = self.filt.read_wgts_file("dummy")
            self.assertEqual(len(result), 27)  # 3 depths * 9 = 27
            self.assertEqual(result.dtype, np.float64)
            npt.assert_array_equal(self.mock_file.read_reals.return_value, result[:9])


    def test_wgts_file_wrong_num_depts(self):
        """
        Testing assertion that Filter.ndep = ndepw
        """
        self.filt.ndep = 5  # Should be 3
        with patch('pytomofilt.filter.spio.FortranFile', return_value=self.mock_file):
            with self.assertRaises(AssertionError):
                self.filt.read_wgts_file("dummy")


    def test_wgts_file_wrong_coefs_per_depth(self):
        """
        Testing assertion that Filter.natd = natd = (lmaxw+1)**2
        """
        self.filt.ndep = 5  # Should be 9
        with patch('pytomofilt.filter.spio.FortranFile', return_value=self.mock_file):
            with self.assertRaises(AssertionError):
                self.filt.read_wgts_file("dummy")


class TestApplyFilter(unittest.TestCase):

    @patch.object(Filter, 'read_eigenvec_file')
    @patch.object(Filter, 'read_wgts_file')
    def setUp(self, mock_read_wgts, mock_read_evec):

        # Setup a dummy Filter object with required attributes
        eigvals = np.array([1.0, 0.4, 0.3, 0.2])
        eigvecs = np.array([[1.0, 0.0, 0.0, 0.0],[0.0, 1.0, 0.0, 0.0],
                            [0.0, 0.0, 1.0, 0.0],[0.0, 0.0, 0.0, 1.0]])
        icrust = 1 
        ismth = 1
        natd = 4
        ndep = 1
        
        mock_read_evec.return_value = (eigvals, eigvecs, icrust,
                                       ismth, natd, ndep)
        mock_read_wgts.return_value = np.arange(1, (natd * ndep) + 1, 
                                                dtype=np.float64)

        self.filt = Filter("dummy_evec", "dummy_wgts", damping=0.1, verbose=False)
    

    def test_apply_filter_size_incorrect(self):
        """
        Test assertion that the input model x has to be larger than 
        the number of eigenvalues and the same dimension as the 
        eigenvectors.
        """
        x_too_small = np.array([0] * 2)
        with self.assertRaises(AssertionError):
            self.filt.apply_filter(x_too_small)

        x_too_large = np.array([0] * 5)
        with self.assertRaises(AssertionError):
            self.filt.apply_filter(x_too_large)


    def test_apply_filter_incorrect_input_datatype(self):
        """
        Test that the elements in input model x must be float64
        objects.
        """
        x_int = np.array([1,2,3,4], dtype=int)
        with self.assertRaises(TypeError):
            self.filt.apply_filter(x_int)
        
        x_32 = np.array([1,2,3,4], dtype=np.float32)
        with self.assertRaises(TypeError):
            self.filt.apply_filter(x_32)


    def test_apply_filter_correctly(self):
        # Create an input vector
        x = np.array([1.0, 2.0, 3.0, 4.0])

        # No model weights or crustal reweighing applied
        self.filt.ismth = 0
        self.filt.icrust = 0
        x_out = self.filt.apply_filter(x)

        self.assertEqual(x_out.shape, x.shape)
        self.assertFalse(np.array_equal(x_out, x))  # Confirm it has been modified
        npt.assert_array_almost_equal(x_out,
                                      np.array([10/11,4/5,3/4,2/3]))

    def test_apply_filter_a_priori_model_weights(self):

        # Create an input vector
        x = np.array([1.0, 2.0, 3.0, 4.0])
    
        # Calculating case with model weight boolean to True
        self.filt.ismth = 1
        self.filt.icrust = 0
        x_model_weights = self.filt.apply_filter(x)

        # Calculating reference case
        self.filt.ismth = 0
        x_ref = self.filt.apply_filter(x)

        npt.assert_array_equal(x_model_weights,
                               x_ref * self.filt.twts)


    def test_apply_filter_crust_weighting(self):

        # Create an input vector
        x = np.array([1.0, 2.0, 3.0, 4.0])
        
        # Calculating case with crustal weighting
        self.filt.icrust = 1
        self.filt.ismth = 0
        x_crustal = self.filt.apply_filter(x)

        # Calculating reference case
        self.filt.icrust = 0
        x_ref = self.filt.apply_filter(x)

        # Check crust weighting has been applied to first natd+1 elements
        npt.assert_array_equal(x_crustal[:self.filt.natd+1],
                               x_ref[:self.filt.natd+1] * 1000.0)


import dataclasses
import os

import numpy as np
import pyshtools as shtools

from . import spline
from . import filter
from . import sh_tools as sh


_rcmb = 3480.0
_rmoho = 6346.691

@dataclasses.dataclass
class RealLayer:
    depth: float
    lats: list[float]
    lons: list[float]
    vals: list[float]


@dataclasses.dataclass
class RealLayerModel:
    layers: list[RealLayer]

def _default_radii(rmin=_rcmb, rmax=_rmoho):
    # Magic numbers from S20RTS model defs (and everything else)
    KNOT_RADII = np.array([1.00000, 0.96512, 0.92675, 0.88454, 0.83810, 0.78701,
                            0.73081, 0.66899, 0.60097, 0.52615, 0.44384, 0.35329,
                            0.25367, 0.14409, 0.02353, -0.10909, -0.25499, -0.41550,
                            -0.59207, -0.78631, -1.00000])
    knots_r = (rmin - rmax) / 2.0 * KNOT_RADII + (rmin + rmax) / 2.0
    return knots_r
    
class RTS_Model:

    def __init__(self, lmax, rmin=_rcmb, rmax=_rmoho, knots=None):
        """
        Create an instance of the RTS model object

        Parameters
        ----------
        lmax: int
            The maximum spherical harmonic degree parameterised in this
            model. 
        rmin: float
            The minimum radii of the RTS_model. If not specified, this
            defaults to the radius of the CMB.
        rmax: float
            The maximum radii of the RTS_model. If not specified, this
            defaults to the radius of the Moho.
        knots: array_like (n)
            The list of n radii points to calculate the model coefficients.
            If None, default knot radii from the RTS models are used, which
            are scaled to the maximum as rmax and minimum as rmin.
        """
        self.lmax = lmax
        self.rmax = rmax
        self.rmin = rmin
        if knots is None:
            # Declare spline knots from default
            self.knots_r = _default_radii(rmin=_rcmb, rmax=_rmoho)
        else:
            self.knots_r = np.asarray(knots) # Should cast into array
        # Build splines
        self.knot_splines = spline.calculate_splines(self.knots_r)
        self.filter_obj = None
        # storage for model - radius, real/imag, degree, order:
        self.coefs = np.zeros((len(self.knots_r), 2, lmax+1, lmax+1))    


    @classmethod
    def from_file(cls, filename, rmin=_rcmb, rmax=_rmoho, knots=None):
        """
        Inputs coefficients from a .sph file. The coefficients should be evaluated
        at the knot points specified. Coefficients are saved under the coefs 
        attribute of this class.

        Parameters
        ----------
        filename : str
            Filename of the .sph file to be read.
        rmin: float
            The minimum radii of the RTS_model. If not specified, this
            defaults to the radius of the CMB.
        rmax: float
            The maximum radii of the RTS_model. If not specified, this
            defaults to the radius of the Moho.
        knots: array_like (n)
            The list of n radii points to calculate the model coefficients.
            If None, default knot radii from the RTS models are used, which
            are scaled to the maximum as rmax and minimum as rmin.
        """
        f = open(filename, 'r')

        header = next(f).split()
        lmax = int(header[0])
        if knots is None:
            knots = _default_radii(rmin=rmin, rmax=rmax)    
        coefs = np.zeros((len(knots), 2, lmax+1, lmax+1))    

        # The format of the rest of the file is a little bit odd. Coefficients 
        # for each radius are listed in blocks and within these blocks coefficients
        # for each l are listed in rows. However, the maximum number of coefficients
        # all of the coefficients for a r,l combination. Things could go wrong if we
        # per line is 11, so for l > 6 we have to read multiple lines before we have
        # end up with too many coefficients for the r,l combination we expect and we
        # trap that at the end of the loop. Otherwise the file could be too short, 
        # in which case we won't have incremented ri enough.
        dataline = []
        ri = 0
        li = 0
        for line in f:  # Reads from second line (header has already been read)
            dataline.extend(line.split())
            
            if len(dataline) == (li * 2) + 1:
                # We have the right number of coefficients for this l
                # so we can process it into the numpy array
                mi = 0
                for m, coef in enumerate(dataline):
                    assert ri < len(knots), f"Too many lines when incrementing ri! ri={ri}"
                    if m == 0:
                        coefs[ri,0,li,mi] = float(coef)
                        mi = mi + 1
                    elif m%2 == 1:
                        # Odd number in list, real coef
                        coefs[ri,0,li,mi] = float(coef)
                        # don't increment mi!
                    else:
                        # even number in list, imag coef
                        coefs[ri,1,li,mi] = float(coef)
                        mi = mi + 1

                li = li + 1

                if li > lmax:
                    # We have read all the coefficients for this radial
                    # layer. Reset L and increment R. 
                    li = 0
                    ri = ri + 1
                dataline = [] # Start adding to a new set of coefficients

            assert len(dataline) < (li * 2) + 1, f"Too much data, li={li}, data={dataline}"

        f.close()
        assert ri == len(knots), f"End of file without seeing all expected radii, ri={ri}, knots={knots}"
        assert li == 0, "End of file without resetting L!"

        # Create new RTS model instance and return
        RTS = cls(lmax, rmin, rmax, knots)
        RTS.coefs = coefs
        return RTS
    

    @classmethod
    def from_terra(self, terra_file, rmin=_rcmb, rmax=_rmoho, knots=None):
        pass

    
    @classmethod
    def from_directory(cls, directory, lmax, rmin=_rcmb, rmax=_rmoho, knots=None):
        """
        Inputs values from files stored in a directory, and automatically reparameterises the data (see 
        reparam for more information). Coefficients are saved under the coefs attribute of this class.
        
        There should be a "depth_layers.dat" file that contains the list of depth layers. Each data file 
        should have an '.dat' extension, numbered in increasing order of depth. The data files be 
        organised into 3 columns, where the columns from left to right are longitude, latitude and the 
        value at the (lon, lat) point respectively (see below):

        Parameters
        ----------
        directory : str
            The directory that contains the files to be read.
        lmax: int
            The maximum spherical harmonic degree parameterised in this
            model. 
        rmin: float
            The minimum radii of the RTS_model. If not specified, this
            defaults to the radius of the CMB.
        rmax: float
            The maximum radii of the RTS_model. If not specified, this
            defaults to the radius of the Moho.
        knots: array_like (n)
            The list of n radii points to calculate the model coefficients.
            If None, default knot radii from the RTS models are used, which
            are scaled to the maximum as rmax and minimum as rmin.
            
        File format example
        ----------
        0.0 90.0 10.0
        0.0 0.0 20.0
        90.0 0.0 20.0
        180.0 0.0 20.0
        270.0 0.0 20.0
        0.0 -90.0 10.0
        
        """

        layers = []
        depth = np.loadtxt(os.path.join(directory, 'depth_layers.dat'))
        data_files = sorted([f for f in os.listdir(directory) if f.endswith('.dat') and f != 'depth_layers.dat'])
        assert len(data_files) == len(depth), "Number of data files does not match number of depths"

        for i, filename in enumerate(data_files):
            data = np.loadtxt(os.path.join(directory, filename))
            layers.append(RealLayer(depth[i], lons = data[:,0],
                                    lats = data[:,1], vals = data[:,2]))
        
        # Write to class
        RTS = cls(lmax, rmin, rmax, knots)
        RTS.reparam(RealLayerModel(layers))
        return RTS
    

    def reparam(self, layer_model):
        """
        Reparameterises the model laterally into spherical harmonics and vertically in a three-point clamped
        cubic spline functions. The output is the coefficients evaluated at the spline knots, and is saved
        in the class s_model attribute.
        
        Parameters
        ----------
        layer_model : RealLayerModel class object
            A list of RealLayer objects, where each RealLayer object represents a list of values at a list
            of longitude and latitude points at a certain depth.
        """
        assert isinstance(layer_model, RealLayerModel), "layer model must be an instance of RealLayerModel"
        
        # Reparameterise file laterally
        layer_depths = [l.depth for l in layer_model.layers]
        sh_coefs = np.zeros((len(layer_depths),) + self.coefs.shape[1:])

        for i, layer in enumerate(layer_model.layers):
            # SHExpandLSQ only works for real coefficients
            cilm, chi2 = shtools.expand.SHExpandLSQ(layer.vals, layer.lats, layer.lons, 
                                                    lmax = self.lmax, norm = 4, csphase = -1)

            # Convert coefficients to RTS format
            sh_coefs[i] = sh.sh_to_rts(cilm)

        # Calculate coefficients at spline knots
        self.coefs = spline.cubic_spline(sh_coefs, layer_depths,
                                         self.knot_splines)


    def filter_from_file(self, evec_file, wght_file, damping, verbose=False):
        """
        Adds a resolution operator (filter) from Fortran formatted files
        
        Parameters
        ----------
        evec_file: file name of the Fortran unformatted file containing the
                   eigenvectors (and values) used to build the resolution
                   operator.
        wght_file: file name of the Fortran unformatted file containing the
                   weights.
        damping: damping parameter used in the inversion.
        verbose: optional, adds extra output for all filter operations
        Setting the optional verbose argument to True results in the more
        output being created. This is useful for debugging and comparing the
        run to the Fortran equivalent
        """
        self.filter_obj = filter.Filter(evec_file, wght_file, damping, verbose)


    def filter(self, model):
        """
        Apply this model's resolution filter to another model instance
        
        Both models must be parameterized the same way (i.e. spherical 
        harmonics at the knots of cubic splines). Typically the model
        with a resolution operator will be from a tomographic inversion
        and the model to be operated on (the argument to this method)
        will be from some other high resolution source - e.g. a geodynamic
        simulation. Note that the filter must be loaded before calling this
        method.
        
        Parameters
        ----------
        model: another instance of the RTS_Model class
        
        Returns
        -------
        filtered_model: RTS_Model class object
            a copy of model, having been filtered using the
            resolution operator. This is the same resolution of the model
            containing the filter.
        """
        assert self.filter_obj is not None, "You must use filter_from_file() to add the filter"
        assert isinstance(model, RTS_Model), "Input model must be an instance of RTS_Model to be filtered" 
        x = model.as_vector()
        x = self.filter_obj.apply_filter(x)
        output_model = RTS_Model(self.lmax, self.rmin, self.rmax, self.knots_r)
        output_model.from_vector(x)
        return output_model
        

    def as_vector(self):
        """
        Return 1D vector representing the model object
        
        This is the order needed for filtering
        (and matches the elements in the files)

        Returns
        -------
        vector: array_like
            a representation of the model coefficients as a 1D vector
        """
        lmx = self.lmax
        ndp = len(self.knots_r)
        natd = (lmx + 1)**2 
        lenatd = natd * ndp
    
        vector = np.empty((lenatd))
        counter = 0
        # FIXME: Maybe numba this loop?
        for ri in range(len(self.knots_r)):
            for li in range(self.lmax + 1):
                for mi in range(li+1):
                    vector[counter] = self.coefs[ri,0,li,mi]
                    counter = counter + 1
                    if mi != 0:
                        vector[counter] = self.coefs[ri,1,li,mi]
                        counter = counter + 1

        assert counter == lenatd, "missing vals" 
        return vector
    

    def from_vector(self, vector):
        """
        Fill out the model coefficients from a 1D vector
        
        Parameters
        -------
        vector: array_like
            a representation of the model coefficients as a 1D vector
        """
        
        counter = 0
        # FIXME: Maybe numba this loop?
        for ri in range(len(self.knots_r)):
            for li in range(self.lmax + 1):
                for mi in range(li+1):
                    self.coefs[ri,0,li,mi] = vector[counter]
                    counter = counter + 1
                    if mi != 0:
                        self.coefs[ri,1,li,mi] = vector[counter]
                        counter = counter + 1


    def correlate(self, model):
        return None

    
    def write(self, filename):
        """
        Writes the coefficients of the model as a .sph file. 

        Parameters
        ----------
        filename : str
            Filename of the .sph file to be saved.
        
        """
        f = open(filename, 'w')

        # Write header - see line 27 of wsphhead.f
        f.write("           {:4} {}{:4} {} \n".format(self.lmax,
                "1"*(self.lmax+1), 24, "000111111111111111111111"))
        # Write the body
        self._write_sph_body_lines(f)
        f.close()


    def _write_sph_body_lines(self, fhandle):
        """
        Write the body of an SPH file to fhandle

        This exists as a seperate method for use in writing SPT files

        Parameters
        ----------
        fhandle : handle
            Handle of the file to be written.
        """
        # Write coefs (knot by knot, and l by l)
        for ri in range(len(self.knots_r)):
            for li in range(self.lmax+1):
                line = ""
                items = 0
                for mi in range(li+1):
                    line = line + " {: 10.4E}".format(self.coefs[ri,0,li,mi])
                    items = items + 1
                    if items == 11: 
                        line = line + "\n"
                        items = 0
                    if mi != 0:
                        # NB: the examples files are all "0.nnnnE-nn" for the 
                        #     values. This gives "n.nnn0E-nn" for the output 
                        #     (with a larger exponent). This is probably OK.
                        #     We could try to get bit for bit equivelance, but
                        #     I struggled with the fortranformat module.
                        line = line + " {: 10.4E}".format(
                                                    self.coefs[ri,1,li,mi])
                        items = items + 1
                        if items == 11: 
                            line = line + "\n"
                            items = 0
                if items != 0:
                    line = line + "\n"
                fhandle.write(line)


import unittest
import numpy as np
import numpy.testing as npt
import tempfile
import os

import pytomofilt.model as mod

class TestDataclass(unittest.TestCase):
    
    def test_dataclass(self):
        """Tests whether RealLayerModel Dataclass saves the correct inputs."""
        # Build a model
        layers = []
        for depth in [10.0, 20.0, 30.0, 50.0, 440.0, 660.0, 1000.0, 2000.0, 2500.0]:
            layers.append(mod.RealLayer(depth,
                                        lats=[90.0, 0.0, 0.0, 0.0, 0.0, -90.0],
                                        lons=[0.0, 0.0, 90.0, 180.0, 270.0, 0.0], 
                                        vals=[10.0, 20.0, 20.0, 20.0, 20.0, 10.0]
                                        ))
            
        layer_model = mod.RealLayerModel(layers)
        self.assertIsInstance(layer_model.layers[1], mod.RealLayer)
        self.assertEqual(layer_model.layers[1].depth, 20.0)
        self.assertEqual(layer_model.layers[1].lats,
                         [90.0, 0.0, 0.0, 0.0, 0.0, -90.0])
        self.assertEqual(layer_model.layers[1].lons,
                         [0.0, 0.0, 90.0, 180.0, 270.0, 0.0])
        self.assertEqual(layer_model.layers[1].vals,
                          [10.0, 20.0, 20.0, 20.0, 20.0, 10.0])


class TestFromFile(unittest.TestCase):
    
    def setUp(self):
        """Create a temporary .sph file for testing."""
        self.test_file = tempfile.NamedTemporaryFile(delete=False, mode='w')
        self.test_file.write("2 1111111111111  24 000111111111111111111111\n")  # lmax = 5
        for i in range(4):
            self.test_file.write("0.1\n")  # Sample coefficients
            self.test_file.write("0.2 0.3 0.4\n")
            self.test_file.write("0.5 0.6 0.7 0.8 0.9\n")
        self.test_file.close()
    

    def tearDown(self):
        """Remove the temporary file after tests."""
        os.unlink(self.test_file.name)
    

    def test_from_file_valid(self):
        """Test normal loading from a valid .sph file."""
        obj = mod.RTS_Model.from_file(self.test_file.name, knots=[0.1, 0.4, 0.6, 0.9])
        self.assertIsInstance(obj, mod.RTS_Model)
        self.assertEqual(obj.lmax, 2)
        self.assertTrue(isinstance(obj.coefs, np.ndarray))
    

    def test_from_file_invalid_format(self):
        """Test loading from an incorrectly formatted file."""
        with tempfile.NamedTemporaryFile(delete=False, mode='w') as bad_file:
            bad_file.write("Invalid Data\n")
            bad_filename = bad_file.name
        
        with self.assertRaises(ValueError):
            mod.RTS_Model.from_file(bad_filename)
        os.unlink(bad_filename)
    

    def test_from_file_missing_file(self):
        """Test behavior when the file does not exist."""
        with self.assertRaises(FileNotFoundError):
            mod.RTS_Model.from_file("non_existent_file.sph")
    

    def test_from_file_empty_file(self):
        """Test behavior when the file is empty."""
        with tempfile.NamedTemporaryFile(delete=False, mode='w') as empty_file:
            empty_filename = empty_file.name
        
        with self.assertRaises(StopIteration):  # Next() call on empty file
            mod.RTS_Model.from_file(empty_filename)
        os.unlink(empty_filename)


    def test_too_many_lines(self):
        """Test behavior when the file has too many lines."""
        with tempfile.NamedTemporaryFile(delete=False, mode='w') as line_file:
            line_file.write("1 1111111111111  24 000111111111111111111111\n")  # lmax = 1
            for i in range(22): # writing for 22 radii instead of 21 (len(knot_radii))
                line_file.write("0.1\n")
                line_file.write("0.2 0.3 0.4\n")
            line_filename = line_file.name
        
        with self.assertRaises(AssertionError):  # Next() call on empty file
            mod.RTS_Model.from_file(line_filename)
        os.unlink(line_filename)


    def test_too_much_data(self):
        """Test behavior when the file has too much data."""
        with tempfile.NamedTemporaryFile(delete=False, mode='w') as data_file:
            data_file.write("2 1111111111111  24 000111111111111111111111\n")  # lmax = 2
            data_file.write("0.1 0.3\n")
            data_filename = data_file.name
        
        with self.assertRaises(AssertionError):  # Next() call on empty file
            mod.RTS_Model.from_file(data_filename)
        os.unlink(data_filename)


class TestFromDirectory(unittest.TestCase):
    
    def setUp(self):
        """Create a temporary directory with sample data."""
        self.test_dir = tempfile.TemporaryDirectory()
        self.depth_file = os.path.join(self.test_dir.name, 'depth_layers.dat')
        with open(self.depth_file, 'w') as f:
            f.write("10.0\n20.0\n30.0\n")
        
        for i in range(3):
            with open(os.path.join(self.test_dir.name, f"layer_{i}.dat"), 'w') as f:
                f.write("0.0 90.0 10.0\n0.0 0.0 20.0\n90.0 0.0 20.0\n")
    
    
    def tearDown(self):
        """Remove the temporary directory after tests."""
        self.test_dir.cleanup()
    

    def test_from_directory_valid(self):
        """Test normal loading from a valid directory."""
        obj = mod.RTS_Model.from_directory(self.test_dir.name, lmax=5, knots=[10,20,30])
        self.assertIsInstance(obj, mod.RTS_Model)
    

    def test_from_directory_missing_depth_file(self):
        """Test behavior when depth_layers.dat is missing."""
        os.remove(self.depth_file)
        with self.assertRaises(FileNotFoundError):
            mod.RTS_Model.from_directory(self.test_dir.name, lmax=5, knots=[10,20,30])
    

    def test_from_directory_no_data_files(self):
        """Test behavior when there are no data files."""
        for i in range(3):
            os.remove(os.path.join(self.test_dir.name, f"layer_{i}.dat"))
        with self.assertRaises(AssertionError):
            mod.RTS_Model.from_directory(self.test_dir.name, lmax=5, knots=[10,20,30])
            

class TestReparam(unittest.TestCase):

    def setUp(self):
        """Set up a test instance and a mock RealLayerModel."""
        self.test_obj = mod.RTS_Model(lmax=5, rmin=1, rmax=10, knots=[10,20,30])
        self.test_obj.coefs = np.zeros((3, 2, 6, 6))

        layers = []
        for depth in [10.0, 20.0, 30.0, 2000.0, 2500.0]:
            layers.append(mod.RealLayer(depth,
                                        lats=[90.0, 0.0, 0.0, 0.0, 0.0, -90.0],
                                        lons=[0.0, 0.0, 90.0, 180.0, 270.0, 0.0], 
                                        vals=[10.0, 20.0, 20.0, 20.0, 20.0, 10.0]
                                        ))
        self.layer_model = mod.RealLayerModel(layers)


    def test_not_layermodel(self):
        """"Test whether layer_model input is a RealLayerModel object"""
        with self.assertRaises(AssertionError):
            self.test_obj.reparam('layer_model') 

    
    def test_reparam_valid(self):
        """Test reparam method with valid input."""
        self.test_obj.reparam(self.layer_model)
        self.assertTrue(isinstance(self.test_obj.coefs, np.ndarray))


class TestWrite(unittest.TestCase):
    
    def setUp(self):
        """Set up a test instance of RTS_Model."""
        self.test_obj = mod.RTS_Model(lmax=5, rmin=1, rmax=10, knots=[0.1, 0.2, 0.3])
        self.test_obj.coefs = np.zeros((3, 2, 6, 6))


    def test_write(self):
        """Test the write method by writing and reading a file."""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            self.test_obj.write(temp_file.name)
            temp_filename = temp_file.name
        
        self.assertTrue(os.path.exists(temp_filename))
        obj = mod.RTS_Model.from_file(temp_filename, knots=[0.1, 0.2, 0.3])
        self.assertEqual(self.test_obj.lmax, obj.lmax)
        npt.assert_array_almost_equal(self.test_obj.coefs, obj.coefs)
        os.remove(temp_filename)


class TestFilterMethods(unittest.TestCase):

    def setUp(self):
        """Set up a test instance of RTS_Model."""
        self.test_obj = mod.RTS_Model(lmax=5, rmin=1, rmax=10, knots=[0.1, 0.2, 0.3])
        self.test_obj.coefs = np.zeros((3, 2, 6, 6))
        self.test_obj.filter_obj = MockFilter()
    

    def test_as_vector(self):
        """Test the as_vector method."""
        vector = self.test_obj.as_vector()
        self.assertEqual(len(vector), (5 + 1) ** 2 * 3)  # Validate vector size
    

    def test_from_vector(self):
        """Test the from_vector method."""
        vector = np.ones((5 + 1) ** 2 * 21)
        self.test_obj.from_vector(vector)
        tril_r = np.tril_indices(6)
        tril_i = tuple(arr+1 for arr in np.tril_indices(5))
        self.assertTrue(np.all(self.test_obj.coefs[:,0,tril_r[0],tril_r[1]] == 1))
        self.assertTrue(np.all(self.test_obj.coefs[:,1,tril_i[0],tril_i[1]] == 1))
    

    def test_filter(self):
        """Test the filter method."""
        model = mod.RTS_Model(lmax=5, rmin=1, rmax=10, knots=[0.1, 0.2, 0.3])
        model.coefs = np.random.rand(3, 2, 6, 6)
        self.test_obj.filter_obj = MockFilter()
        filtered_model = self.test_obj.filter(model)
        self.assertIsInstance(filtered_model, mod.RTS_Model)
    

    def test_filter_without_filter_obj(self):
        """Ensure filter raises an error if filter_obj is not set."""
        self.test_obj.filter_obj = None
        model = mod.RTS_Model(lmax=5, rmin=1, rmax=10, knots=[0.1, 0.2, 0.3])
        with self.assertRaises(AssertionError):
            self.test_obj.filter(model)
    

    def test_filter_invalid_input(self):
        """Ensure filter raises an error for invalid input type."""
        with self.assertRaises(AssertionError):
            self.test_obj.filter("invalid_input")
    

class MockFilter:
    def apply_filter(self, vector):
        return vector * 0.5  # Mock filtering behavior


import scipy.interpolate as spi
import scipy.linalg as spl
import numpy as np

# Function to calculate 3-point derivative clamped spline
def clamped_cubic_spline(x, y):
    """ 
    Creates a Scipy CubicSpline object with a 3-point clamped spline condition.
    The CubicSpline object can take an argument of a value x and output the value
    y of the spline at x.

    Parameters
    ----------
    x : array_like (n)
        1-D array containing values of the independent variable. Values must be real,
        finite and in strictly increasing order.
    y : array_like (n, k)
        Array containing values of the dependent variable. The length along the first 
        axis must match the length of x. Values must be finite.

    Returns
    -------
    scipy.interpolate.CubicSpline object
        A callable object that returns the value of the splines at the called location.
    """
    # Check that x must at least have 3 entries
    assert len(x) >= 3, "x must have 3 or more entries"
    # Check that entries in x are in ascending order
    assert np.all(sorted(x) == x), "List must be in ascending order" 
    # Check that length of x == length of y at axis = 0
    assert len(x) == y.shape[0], "length of y at axis 0 must be the same length as x"

    start_deriv = (y[0]*(2*x[0]-x[1]-x[2])*(x[2]-x[1]) + y[1]*(x[2]-x[0])**2 - y[2]*(x[1] - x[0])**2)/((x[1] - x[0]) * (x[2] - x[0]) * (x[2] - x[1]))
    end_deriv = (y[-1]*(2*x[-1]-x[-2]-x[-3])*(x[-3]-x[-2]) + y[-2]*(x[-3]-x[-1])**2 - y[-3]*(x[-2] - x[-1])**2)/((x[-2] - x[-1]) * (x[-3] - x[-1]) * (x[-3] - x[-2]))
    ccs = spi.CubicSpline(x, y, bc_type=((1, start_deriv),(1, end_deriv)), extrapolate=None)
    
    return ccs

# Cubic interpolation on spline knots (21)
def calculate_splines(knots):
    """
    Calculate_splines calculates the cubic splines based on the knot points provided

    Parameters
    ----------
    knots : list (n)
        A list of locations at which to anchor the splines at. Entries need to be in
        ascending order

    Returns
    -------
    scipy.interpolate.CubicSpline object
        A callable object that returns the weights of each of the splines at the called location.
    """
    spl = np.identity(len(knots))
    return clamped_cubic_spline(knots,spl)

def cubic_spline(coefs, depth, splines):
    """
    Cubic_spline calculates coefficients at the knot points

    Parameters
    ----------
    coefs : array (n,)
        An array containing the coefficients evaluated at each depth. The first axis
        must be the same length as depth.
    
    depth: list (n)
        A list containing values of the depth of each layer.

    splines: scipy.interpolate.CubicSpline object
        A callable CublineSpline object, which is the output from the calculate_splines
        function above. It returns the weights of each of the splines at the called 
        location.

    Returns
    -------
    coef_rts: np.ndarray
        An array of coefficients evaluated at each of the spline knots.
    """

    assert len(depth) == coefs.shape[0], "first axis of coefs array must be the \
                                          same length as depth"
    assert len(depth) >= len(splines.x), f"Number of depth points must be equal \
                                          or more than {len(splines.x)}"

    if coefs.ndim > 2:
        coefs_redim = coefs.reshape((len(depth), -1))
    else:
        coefs_redim = coefs

    wgt_layers = splines(depth)
    coef_rts = spl.lstsq(wgt_layers,coefs_redim)[0]

    if coef_rts.ndim == 1:
        coef_rts = coef_rts[:,None] # mandating that matrix must be 2 dimensional
    elif coefs.ndim > 2:
        coef_rts = coef_rts.reshape((wgt_layers.shape[-1],) + coefs.shape[1:])

    return coef_rts


import os

import numpy as np
import scipy.io as spio
import numba

class Filter(object):
    """
    Object to apply tomographic resolution operator to seismic model

    The creation of tomographic models of seismic wave velocities 
    such as S20RTS, S40RTS and SP12RTS can also provide a resolution
    operator that maps a true earth model into the tomography
    model, and fully describes the spatially heterogeneous resolution
    of the tomography. This object allows a resolution operator from
    the "RTS" family of models to be applied to a model of seismic
    velocities, such that the velocities can be fairly compared with
    the tomography. Software for doing this also exists in Fortran.
    The main difference between this implementation and the Fortran
    version is that this version holds the operator in memory allowing
    multiple models to be "filtered". The Fortran implementation reads
    applies one "line" of the operator at a time reading the line from
    disk. This means that this implementation uses much more memory 
    but is much faster if multiple models have to be filtered. On the
    other hand, the Fortran implementation can be used on low memory 
    machines. Within LEMA where we make use of very large numbers of
    low resolution models holding the model in memory is clearly the
    correct choice (indeed, it is ~100 times faster to do it this way).

    To use the filter first create an instance of this class while
    providing the file names of the files containing the resolution
    operator (this is quite slow, see the documentation of the __init__
    method). Then convert your model of seismic velocities into an
    RTSJointShell object before calling the RTSJointShell.filter_sp_model
    method with the instance provided in the "filter" keyword. The 
    RTSJointShell object is then updated to represent the "filtered"
    seismic velocities.
    """
    
    def __init__(self, evec_file, wght_file, damping, verbose=False):
        """
        Create an instance of the resolution operator object

        This needs two files and a damping parameter which are passed
        in as the following arguments:

        evec_file: file name of the Fortran unformatted file containing the
                   eigenvectors (and values) used to build the resolution
                   operator.
        wght_file: file name of the Fortran unformatted file containing the
                   weights;
        damping: damping parameter used in the inversion

        Setting the optional verbose argument to True results in the more
        output being created. This is useful for debugging and comparing the
        run to the Fortran equivalent
        """
        self.evcfl = evec_file
        self.wgtsfl = wght_file
        self.damp = damping
        self.verbose = verbose
        
        # Read weights and eigenvector files
        (self.eigvals, self.eigvecs, self.icrust, 
         self.ismth, self.natd, self.ndep) = self.read_eigenvec_file(self.evcfl)
        self.twts = self.read_wgts_file(self.wgtsfl)

    def read_eigenvec_file(self, filename):
        """
        Read evec file for SP12RTS filtering
        
        This contains the eigenvectors and eigenvalues
        of the model resolution matrix. The file is an
        unformatted fortran file, so we read using 
        scipy.io.FortranFile and note that this format
        does not form part of any standard so is (Fortran)
        processor dependent. We rely on the format being
        readable to us. There are three ways of terminating
        reading. (1) We have read all lenatd eigenvectors,
        (2) the eigenvalues (which are sorted) is smaller
        that a (damping dependent value) or (3) fewer than
        this number of vectors are in the file.
        """
    
        f = spio.FortranFile(filename, 'r')
        
        # Header line 1
        (lmaxh, numatd2, ndep, icrust, idensi1, idum1,
            ismth ) = f.read_ints()
        # Header line 2
        record = f.read_ints()
        mp1 = record[0]
        iparsw = record[1:mp1+1] # What is in the file
        parwts = record[mp1+1:2*mp1+1] # Unclear, only zeros
        # Next bit is a matrix of number of spherical harmonics
        # and number of something else for each part of file.
        # note we need to change order on read and transpose
        # as FortranFile doesn't deal with the different 
        # array order.
        ipardps = record[2*mp1+1:4*mp1+1].reshape((mp1,2)).T 

        # The Fortran version uses data from the spt file to work
        # out how much to read. We use the same information from 
        # the eigenvector file (and check things match before we
        # do any filtering).
        natd = (lmaxh + 1)**2
        assert natd == numatd2, "header size mismatch"
        lenatd = natd * ndep
        
        # Read in all the eigenvectors
        eigvals = np.empty((lenatd))
        eigvecs = np.empty((lenatd,lenatd))
        
        # Read first value and vector record
        # in order to set stop condition
        record = f.read_reals()
        eigvals[0] = record[0]
        eigvecs[0,:] = record[1:lenatd+1]
        eta=eigvals[0] * self.damp
        if self.verbose:
            print("Largest eigenvalue:", eigvals[0], 
                  "ets:", eta)
        stpfact=eta/5000.

        # Read rest of eigen file
        for i in range(1, lenatd+1):
            try:
                record = f.read_reals()
            except:
                if self.verbose:
                    print("Reduced EVC file?")
                    print(i, "vectors read")
                i = i - 1 # Don't use this one
                break
                
            eigvals[i] = record[0]
            eigvecs[i,:] = record[1:lenatd+1]
    
            if eigvals[i] < stpfact:
                if self.verbose:
                    print("Stop building model")
                    print("Last Eigenvector used is", i)
                break
                
        f.close()
                
        return eigvals[:i+1], eigvecs[:i+1, :], icrust, ismth, natd, ndep
    
    
    def read_wgts_file(self, filename):
        """
        Read the weights file
        
        Again unformatted. One interesting feature is that the
        type does not seem to be real*8, reads OK as integer...
        ... so I assume that integer (default prec) is OK and 
        then cast to a float in python (implicit type of twts
        should be real).
        """
        f = spio.FortranFile(filename, 'r')
        (lmaxw, nsmn, nsmx, ndepw, etaz, etah,
          etai, iderh, iderv) = f.read_ints()
        assert ndepw == self.ndep, \
            "Number of depts in wgts and eigv files differ"
        natd = (lmaxw + 1)**2
        assert natd == self.natd, \
            "Number of coefs per depth in wgts and eigv files differ"
        lenatd = natd * ndepw
        twts = np.empty((lenatd))
        for i in range(ndepw):
            ind = (i) * natd
            
            record = f.read_reals('f4')
            
            twts[ind:ind+natd] = record.astype(np.float64)
        
        f.close()

        return twts
    
    def apply_filter(self, x):
        """
        Applies the tomography filter
        
        Reimplementation of the guts of mk3d_res_ESEP.f
        but with array operations when this is easy.
        
        Arguments: x. The input model as a 1D vector. If the 
        model is represented by a set of values varying in radius,
        r, degree, l and order, m, with real and imaginary components the 
        order of the elements (r, l, m, s) in the vector is given by:
        (0, 0, 0, r), (0, 1, 0, r), (0, 1, 1, r), (0, 1, 1, i), 
        (0, 2, 0, r), (0, 2, 1, r), (0, 2, 1, i), (0, 2, 2, r),
        (0, 2, 2, i), (0, 3, 0, r), (0, 3, 1, r), (0, 3, 1, i)
        ... (0, lmax, lamx, r), (0, lmax, lmax, i) ... 
        (rmax, lmax, lmax, r), (rmax, lmax, lmax, i). That is, 
        imaginary coeffs are skipped when they do not exist and the
        data increments through l, m, then r, in that order.
 
        Returns: xout. The filtered model (same structure as x)
        """
        assert x.size >= self.eigvals.size, \
            "Model size / filter mismatch (eval)"
        assert x.size == self.eigvecs[0,:].size, \
            "Model size / filter mismatch (eec)"
        x_out = apply_filter_inner(x, self.eigvals, self.eigvecs, self.twts, self.damp)
                
        if self.ismth == 1:
            if self.verbose:
                print("applying a priori model weights")
            x_out = x_out * self.twts
            
        if self.icrust == 1:
            if self.verbose:
                print("Reweight the crustal thickness")
            x_out[0:self.natd+1] = x_out[0:self.natd+1] * 1000.0
            
        return x_out
    

# jit compiling this bit makes the time per loop go from
# 7.56 s to 59.6 ms, running it in parallel gives us another 
# order of magnitude speedup. This gets hidden by IO if we 
# only filter one model. Note the parallel reduction 
# over x_out and the prange function.
@numba.jit('f8[:](f8[:],f8[:],f8[:,:],f8[:],f8)', nopython=True, parallel=True)
def apply_filter_inner(x, eigvals, eigvecs, twts, damp):
    x_out = np.zeros_like(x)
    twtsinv = 1.0 / twts
    eta = eigvals[0] * damp
    # loop over the eigenvectors and values 
    # we read in
    for i in numba.prange(eigvals.size):
            
        w = ( (eigvals[i] / (eigvals[i] + eta)) * 
               np.sum(twtsinv[:] * x[:] * eigvecs[i,:]) )

        x_out += w * eigvecs[i,:]
            
    return x_out




import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import pyshtools

from . import sh_tools as sh

def plot_shcoefs(coefs, r=None, title='', quantity='', cmap='RdBu',
                      coast_color='k', projection=ccrs.Robinson(),
                      scale_factor=None, lmax=None, levels=None, vmin=None,
                      vmax=None, extend='neither', fig=None, ax=None):
    """
    Plot a set of spherical harmonic coefficients `coef`.  Label the radius with `r` (km)
    and the value with `quantity`.  Harmonics are truncated at degree `lmax`.
    If `fig` and `ax` are supplied, then the plot is added to the provided matplotlib
    figure and axis handles, respectively.
    """

    assert ( (ax is not None) and (fig is not None) or
             (ax is     None) and (fig is     None)), "Arg error: Axis set, but fig is None"

    if lmax is None:
        assert coefs.shape[1] == coefs.shape[2], "SH coefficients are not the correct shape"
        lmax = coefs.shape[1] - 1

    # For very "clean" coefficents (i.e. many zeros) the contour function
    # messes up and we get big chunks of the map with nothing plotted.
    # Avoid this by adding a small value to all coefficents.
    # See issue 42 on github.
    coefs = coefs + 1E-15

    # Convert coefs from RTS to SH format
    sh_coefs = sh.rts_to_sh(coefs)

    # Set plot area - fixed for now but we could use this to look at a local
    # area
    north = 90.0
    south = -90.0
    west = 0.0
    east = 360.0
    interval = 3.0
    # NB - arange goes from low to high, but we need N to S so reverse output.
    lats = np.arange(south, north+interval, interval)[::-1]
    lons = np.arange(west, east+interval, interval)
    grid = pyshtools.expand.MakeGrid2D(sh_coefs, interval, lmax, norm=4, north=north, south=south,
               east=east, west=west)
    if scale_factor is not None:
        grid = grid * scale_factor
    lons, lats = np.meshgrid(lons, lats)

    if ax is None and fig is None:
        fig, ax = plt.subplots(subplot_kw=dict(projection=projection),
                           figsize=(15,5))

    ax.set_global() # wrapping
    # No idea why, but we need "PlateCarree" here even if we use something
    # else (e.g. Robinson) for the projection above. This is odd but see
    # http://scitools.org.uk/cartopy/docs/v0.5/matplotlib/introductory_examples/
    # ... 03.contours.html
    h = ax.contourf(lons, lats, grid, 100,
                    transform=ccrs.PlateCarree(),
                    cmap=cmap, levels=levels, vmin=vmin, vmax=vmax, extend=extend)

    ax.coastlines(color=coast_color)
    if projection == ccrs.PlateCarree():
        ax.set_xticks(np.linspace(-180, 180, 5), crs=projection)
        ax.set_yticks(np.linspace(-90, 90, 5), crs=projection)
    
    ax.tick_params(axis='both', which='major', labelsize=15)
    cb = fig.colorbar(h, ax=ax, orientation='vertical', fraction=0.02)
    cb.set_label(quantity, size=15)
    cb.ax.tick_params(axis='both', which='major', labelsize=15)

    if r is None:
        ax.set_title('{}'.format(title), size=15)
    else:
        ax.set_title('{} at {} km'.format(title, r), size=15)


import unittest
import numpy as np
import numpy.testing as npt

import pytomofilt.spline as spl

class TestClampedCubicSpline(unittest.TestCase):

    def test_less_than_three(self):
        """Test whether the spline has more than 3 points."""
        with self.assertRaises(AssertionError):
            spl.clamped_cubic_spline([1,2], [1,2])


    def test_not_ascending_order(self):
        """Test whether the x input for the spline is increasing."""
        with self.assertRaises(AssertionError):
            spl.clamped_cubic_spline([1,5,6,2,3], [1,2,3,4,5])


    def test_different_x_y_length(self):
        """Test for equal length of inputs x and y"""
        x = np.array([1,2,3,4,5])
        y = np.array([[3,2,1,5], [1,2,3,4]])

        with self.assertRaises(AssertionError):
            spl.clamped_cubic_spline(x,y)


    def test_spline(self):
        """
        Test that the spline contains the correct properties and
        returns the correct values
        """
        x = np.array([1,2,3,4,5])
        y = np.array([[1,2,3,4,5], [5,2,3,1,0]]).T
        ccs = spl.clamped_cubic_spline(x,y)

        npt.assert_array_equal(ccs.x, x)
        npt.assert_array_equal(ccs(x[0]), [1,5])


class TestCalculateSpline(unittest.TestCase):
    
    def __init__(self, *args, **kwargs):
        """Initialises an example cubic spline object."""
        super(TestCalculateSpline, self).__init__(*args, **kwargs)
        
        # Create a CubicSpline object
        knots = np.array([1,2,6,8,10.0,13.2,15.7,18.0])
        self.splines = spl.calculate_splines(knots)


    def test_knot_points(self):
        """"Test that the spline gives the right weighting at a knot point."""
        npt.assert_array_equal(self.splines(8), [0,0,0,1,0,0,0,0])
    
    
    def test_sum_to_one(self):
        """Test that the spline is normalised to 1."""
        self.assertEqual(sum(self.splines(3.1235)), 1)


class TestCubicSpline(unittest.TestCase):
    
    def __init__(self, *args, **kwargs):
        """Initialises an example cubic spline object."""
        super(TestCubicSpline, self).__init__(*args, **kwargs)
        
        # Create a CubicSpline object to use for testing cubic_spline function
        self.knots = np.arange(0,22,3)
        self.splines = spl.calculate_splines(self.knots)
    

    def test_unequal_depth_and_coefs(self):
        """Test that the first dimension of coefs and depth inputs must be equal."""
        with self.assertRaises(AssertionError):
            spl.cubic_spline(np.array([0.6,0.8,0.2]), 
                                      [1,2,3,4], self.splines)


    def test_less_depth_points_than_knots(self):
        """
        Test that the length of the depth input must be more than the number
        of knot points.
        """
        with self.assertRaises(AssertionError):
            # number of depths (5) < number of knots (8)
            spl.cubic_spline(np.array([0.2,0.6,0.1,-0.3,-1]), 
                             [1,2,3,4,5], self.splines) 


    def test_same_knots_and_depth(self):
        """
        Test that the input and output coefficients should be the same if the
        input depth values are the same as the knot points.
        """
        pts = self.knots
        spl_mtx = np.array([[1,2],[3,5],[5,3.2],[2,1.5],
                            [3,2.2],[5.5,5.5],[2.3,0.3],[8.0,-3.0],])
        coefs = spl.cubic_spline(spl_mtx, pts, self.splines)

        # Accurate to 10 decimal places
        npt.assert_array_almost_equal(coefs, spl_mtx, 10) 


    def test_return_correct_coefs(self):
        """
        Test that spl.cubic_spline returns the correct coefficients for
        an example set of coefficients and depth inputs.
        """
        coefs_at_knots = np.array([5,6.7,8.0,1.5,2,3.4,5,0.2])

        # Forward calculation to calculate coefficients at points
        pts = np.array([1,3.5,6,7.2,8.2,9.3,10,15,16.2,25])
        spl_mtx = np.matmul(self.splines(pts), coefs_at_knots)
        coefs_calculated = spl.cubic_spline(spl_mtx, pts, self.splines)

        # Accurate to 10 decimal places
        npt.assert_array_almost_equal(coefs_at_knots, 
                                      coefs_calculated.ravel(), 10) 